/* Amiga Keyboard driver code ripped from Linux 2.6.0 kernel */

/*
 * $Header: /code/cw/device/cwkey.c,v 1.3 2004/02/11 16:35:14 doj Exp $
 *
 *  Copyright (c) 2000-2001 Vojtech Pavlik
 *
 *  Based on the work of:
 *	Hamish Macdonald
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Should you need to contact me, the author, you can do so either by
 * e-mail - mail your message to <vojtech@ucw.cz>, or by paper mail:
 * Vojtech Pavlik, Simunkova 1594, Prague 8, 182 00 Czech Republic
 */

#include "cw.h"

#include <asm/io.h>
#include <linux/input.h>

static unsigned char amikbd_keycode[0x78] = {
  /* 0x00 */ 41,  2,  3,  4,  5,  6,  7,  8,
  /* 0x08 */ 9,  10, 11, 12, 13, 43,  0, 82,
  /* 0x10 */ 16, 17, 18, 19, 20, 21, 22, 23,
  /* 0x18 */ 24, 25, 26, 27,  0, 79, 80, 81,
  /* 0x20 */ 30, 31, 32, 33, 34, 35, 36, 37,
  /* 0x28 */ 38, 39, 40, 43,  0, 75, 76, 77,
  /* 0x30 */ 86, 44, 45, 46, 47, 48, 49, 50,
  /* 0x38 */ 51, 52, 53,  0, 83, 71, 72, 73,
  /* 0x40 */ 57, 14, 15, 96, 28,  1,111,  0,
  /* 0x48 */ 0,  0,  74,  0,103,108,106,105,
  /* 0x50 */ 59, 60, 61, 62, 63, 64, 65, 66,
  /* 0x58 */ 67, 68,179,180, 98, 55, 78,138,
  /* 0x60 */ 42, 54, 58, 29, 56,100,125,126
};

static const char *amikbd_messages[8] = {
  KERN_ALERT "amikbd: Ctrl-Amiga-Amiga reset warning!!\n",
  KERN_WARNING "amikbd: keyboard lost sync\n",
  KERN_WARNING "amikbd: keyboard buffer overflow\n",
  KERN_WARNING "amikbd: keyboard controller failure\n",
  KERN_ERR "amikbd: keyboard selftest failure\n",
  KERN_INFO "amikbd: initiate power-up key stream\n",
  KERN_INFO "amikbd: terminate power-up key stream\n",
  KERN_WARNING "amikbd: keyboard interrupt\n"
};

static struct input_dev amikbd_dev[CWSID_MAX_NUM];

static char *amikbd_name = "Amiga keyboard";

void amikbd_poll(void)
{
  int i;
  for(i=0; i<cw_num; i++)
    // check for new key
    if(inb(cw_register[i][KeyStatus])&0x80)
      {
	unsigned char scancode, down;

	scancode=inb(cw_register[i][KeyDat]);
	down = !(scancode&128);
	scancode&=127;
#if 0
	printk(KERN_INFO "scancode: %02x -> %02x\n", scancode, amikbd_keycode[scancode]);
#endif
	/* scancodes < 0x78 are keys */
	if (scancode < 0x78)
	  {
	    scancode = amikbd_keycode[scancode];

	    /* CapsLock is a toggle switch key on Amiga */
	    if (scancode == KEY_CAPSLOCK)
	      {	
		input_report_key(&amikbd_dev[i], scancode, 1);
		input_report_key(&amikbd_dev[i], scancode, 0);
	      }
	    else
	      input_report_key(&amikbd_dev[i], scancode, down);

	    input_sync(&amikbd_dev[i]);
	  }
	/* scancodes >= 0x78 are error codes */
	else
	  printk(amikbd_messages[scancode - 0x78]);
	outb(0, cw_register[i][KeyDat]);
      }
}

static int initialized=0;

int __init amikbd_init(void)
{
  static char str[CWSID_MAX_NUM][32];
  int c;
  for(c=0; c<cw_num; c++)
    {
      int i;

      init_input_dev(&amikbd_dev[c]);

      amikbd_dev[c].evbit[0] = BIT(EV_KEY) | BIT(EV_REP);
      amikbd_dev[c].keycode = amikbd_keycode;
      amikbd_dev[c].keycodesize = sizeof(unsigned char);
      amikbd_dev[c].keycodemax = ARRAY_SIZE(amikbd_keycode);

      for (i = 0; i < 0x78; i++)
	if (amikbd_keycode[i])
	  set_bit(amikbd_keycode[i], amikbd_dev[c].keybit);

      amikbd_dev[c].name = amikbd_name;
      sprintf(str[c], "CatWeasel %i Keyboard", c);
      amikbd_dev[c].phys = str[c];
      amikbd_dev[c].id.bustype = BUS_AMIGA;
      amikbd_dev[c].id.vendor = 0x0001;
      amikbd_dev[c].id.product = 0x0001;
      amikbd_dev[c].id.version = 0x0100;

      input_register_device(&amikbd_dev[c]);

      /* enable keyboard */
      outb(0, cw_register[c][KeyStatus]);
      outb(0, cw_register[c][KeyDat]);
    }

  initialized=1;
  return 0;
}

void __exit amikbd_exit(void)
{
  if(initialized)
    {
      int c;
      for(c=0; c<cw_num; c++)
	input_unregister_device(&amikbd_dev[c]);
      initialized=0;
    }
}
