/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>
  The 6510 emulation code is (c) 1999 by Dirk Jadgmann and Tammo Hinrichs

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /code/cw/sid/6510.cc,v 1.11 2004/05/14 14:51:03 doj Exp $
*/

#include "6510.h"
#include "c64.h"
#include "defines.h"
#include "stilinfo.hpp"

static char *disOP[]= {
  "adc","and","asl","bcc","bcs","beq","bit","bmi","bne","bpl","brk","bvc","bvs","clc",
  "cld","cli","clv","cmp","cpx","cpy","dec","dex","dey","eor","inc","inx","iny","jmp",
  "jsr","lda","ldx","ldy","lsr","nop","ora","pha","php","pla","plp","rol","ror","rti",
  "rts","sbc","sec","sed","sei","sta","stx","sty","tax","tay","tsx","txa","txs","tya",
  "*aso/slo", "*rla", "*lse/sre", "*rra", "*axs/sax", "*lax", "*dcm/dcp", "*ins/isc",
  "*alr", "*arr", "*xaa", "*oal", "*sax", "*skb", "*skw", "*hlt",
  "*tas", "*say", "*xas", "*axa", "*anc", "*las"
};

static char *disADR1[]={"", "#", "", "", "", "", "", "", "(", "(", "(", "", "r", ""};
static char *disADR2[]={
  "           ",
  "           ",
  "           ",
  ",x         ",
  ",y         ",
  "           ",
  ",x         ",
  ",y         ",
  ")          ",
  ",x)        ",
  "),y        ",
  "           ",
  "           ",
  "           "
};

static int disADRl[]={0,1,2,2,2,1,1,1,2,1,1,0,1,0};

static opcode_t opcodes[256]= {
  /* 00 */ BRK,ORA,HLT,ASO,SKB,ORA,ASL,ASO,
  /* 08 */ PHP,ORA,ASL,ANC,SKW,ORA,ASL,ASO,
  	   
  /* 10 */ BPL,ORA,HLT,ASO,SKB,ORA,ASL,ASO,
  /* 18 */ CLC,ORA,NOP,ASO,SKW,ORA,ASL,ASO,
  	   
  /* 20 */ JSR,AND,HLT,RLA,BIT,AND,ROL,RLA,
  /* 28 */ PLP,AND,ROL,ANC,BIT,AND,ROL,RLA,
  	   
  /* 30 */ BMI,AND,HLT,RLA,SKB,AND,ROL,RLA,
  /* 38 */ SEC,AND,NOP,RLA,SKW,AND,ROL,RLA,
  	   
  /* 40 */ RTI,EOR,HLT,LSE,SKB,EOR,LSR,LSE,
  /* 48 */ PHA,EOR,LSR,ALR,JMP,EOR,LSR,LSE,
  	   
  /* 50 */ BVC,EOR,HLT,LSE,SKB,EOR,LSR,LSE,
  /* 58 */ CLI,EOR,NOP,LSE,SKW,EOR,LSR,LSE,
  	   
  /* 60 */ RTS,ADC,HLT,RRA,SKB,ADC,ROR,RRA,
  /* 68 */ PLA,ADC,ROR,ARR,JMP,ADC,ROR,RRA,
  	   
  /* 70 */ BVS,ADC,HLT,RRA,SKB,ADC,ROR,RRA,
  /* 78 */ SEI,ADC,NOP,RRA,SKW,ADC,ROR,RRA,
  	   
  /* 80 */ SKB,STA,SKB,AXS,STY,STA,STX,AXS,
  /* 88 */ DEY,SKB,TXA,XAA,STY,STA,STX,AXS,
  	   
  /* 90 */ BCC,STA,HLT,AXA,STY,STA,STX,AXS,
  /* 98 */ TYA,STA,TXS,TAS,SAY,STA,XAS,AXA,
  	   
  /* a0 */ LDY,LDA,LDX,LAX,LDY,LDA,LDX,LAX,
  /* a8 */ TAY,LDA,TAX,OAL,LDY,LDA,LDX,LAX,
  	   
  /* b0 */ BCS,LDA,HLT,LAX,LDY,LDA,LDX,LAX,
  /* b8 */ CLV,LDA,TSX,LAS,LDY,LDA,LDX,LAX,
  	   
  /* c0 */ CPY,CMP,SKB,DCM,CPY,CMP,DEC,DCM,
  /* c8 */ INY,CMP,DEX,SAX,CPY,CMP,DEC,DCM,
  	   
  /* d0 */ BNE,CMP,HLT,DCM,SKB,CMP,DEC,DCM,
  /* d8 */ CLD,CMP,NOP,DCM,SKW,CMP,DEC,DCM,
  	   
  /* e0 */ CPX,SBC,SKB,INS,CPX,SBC,INC,INS,
  /* e8 */ INX,SBC,NOP,SBC,CPX,SBC,INC,INS,
  	   
  /* f0 */ BEQ,SBC,HLT,INS,SKB,SBC,INC,INS,
  /* f8 */ SED,SBC,NOP,INS,SKW,SBC,INC,INS,
};

static addrmode_t modes[256]= {
  /* 00 */ IMP,INDX,IMP,INDX,ZP,ZP,ZP,ZP,
  /* 08 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* 10 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPX,ZPX,
  /* 18 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSX,ABSX,
  	   
  /* 20 */ ABS,INDX,IMP,INDX,ZP,ZP,ZP,ZP,
  /* 28 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* 30 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPX,ZPX,
  /* 38 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSX,ABSX,
  	   
  /* 40 */ IMP,INDX,IMP,INDX,ZP,ZP,ZP,ZP,
  /* 48 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* 50 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPX,ZPX,
  /* 58 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSX,ABSX,
  	   
  /* 60 */ IMP,INDX,IMP,INDX,ZP,ZP,ZP,ZP,
  /* 68 */ IMP,IMM,ACC,IMM,IND,ABS,ABS,ABS,
  	   
  /* 70 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPX,ZPX,
  /* 78 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSX,ABSX,
  	   
  /* 80 */ IMM,INDX,IMM,INDX,ZP,ZP,ZP,ZP,
  /* 88 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* 90 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPY,ZPY,
  /* 98 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSY,ABSY,
  	   
  /* a0 */ IMM,INDX,IMM,INDX,ZP,ZP,ZP,ZP,
  /* a8 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* b0 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPY,ZPY,
  /* b8 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSY,ABSY,
  	   
  /* c0 */ IMM,INDX,IMM,INDX,ZP,ZP,ZP,ZP,
  /* c8 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* d0 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPX,ZPX,
  /* d8 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSX,ABSX,
  	   
  /* e0 */ IMM,INDX,IMM,INDX,ZP,ZP,ZP,ZP,
  /* e8 */ IMP,IMM,ACC,IMM,ABS,ABS,ABS,ABS,
  	   
  /* f0 */ REL,INDY,IMP,INDY,ZPX,ZPX,ZPX,ZPX,
  /* f8 */ IMP,ABSY,ACC,ABSY,ABSX,ABSX,ABSX,ABSX,
};

const byte MOS6510::getaddr(const int mode)
{
  word ad,ad2;
  switch(mode)
    {
    case IMP:
      cyclesSingle+=2;
      return 0;
    case IMM:
      cyclesSingle+=2;
      return comp->getmem(pc++);
    case ABS:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad|=comp->getmem(pc++)<<8;
      return comp->getmem(ad);
    case ABSX:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad|=256*comp->getmem(pc++);
      ad2=ad+x;
      if ((ad2&0xff00)!=(ad&0xff00))
        cyclesSingle++;
      return comp->getmem(ad2);
    case ABSY:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad|=256*comp->getmem(pc++);
      ad2=ad+y;
      if ((ad2&0xff00)!=(ad&0xff00))
        cyclesSingle++;
      return comp->getmem(ad2);
    case ZP:
      cyclesSingle+=3;
      ad=comp->getmem(pc++);
      return comp->getmem(ad);
    case ZPX:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad+=x;
      return comp->getmem(ad&0xff);
    case ZPY:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad+=y;
      return comp->getmem(ad&0xff);
    case INDX:
      cyclesSingle+=6;
      ad=comp->getmem(pc++);
      ad+=x;
      ad2=comp->getmem(ad&0xff);
      ad++;
      ad2|=comp->getmem(ad&0xff)<<8;
      return comp->getmem(ad2);
    case INDY:
      cyclesSingle+=5;
      ad=comp->getmem(pc++);
      ad2=comp->getmem(ad);
      ad2|=comp->getmem((ad+1)&0xff)<<8;
      ad=ad2+y;
      if ((ad2&0xff00)!=(ad&0xff00)) // this must be wrong
        cyclesSingle++;
      return comp->getmem(ad);
    case ACC:
      cyclesSingle+=2;
      return a;
    }
  return 0;
}

void MOS6510::setaddr(const int mode, const byte val)
{
  word ad,ad2;
  switch(mode)
    {
    case ABS:
      cyclesSingle+=2;
      ad=comp->getmem(pc-2);
      ad|=256*comp->getmem(pc-1);
      comp->setmem(ad,val);
      return;
    case ABSX:
      cyclesSingle+=3;
      ad=comp->getmem(pc-2);
      ad|=256*comp->getmem(pc-1);
      ad2=ad+x;
      if ((ad2&0xff00)!=(ad&0xff00))
        cyclesSingle--;
      comp->setmem(ad2,val);
      return;
    case ZP:
      cyclesSingle+=2;
      ad=comp->getmem(pc-1);
      comp->setmem(ad,val);
      return;
    case ZPX:
      cyclesSingle+=2;
      ad=comp->getmem(pc-1);
      ad+=x;
      comp->setmem(ad&0xff,val);
      return;
    case ACC:
      a=val;
      return;
    }
}

void MOS6510::putaddr(const int mode, const byte val)
{
  word ad,ad2;
  switch(mode)
    {
    case ABS:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad|=comp->getmem(pc++)<<8;
      comp->setmem(ad,val);
      return;
    case ABSX:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad|=comp->getmem(pc++)<<8;
      ad2=ad+x;
      comp->setmem(ad2,val);
      return;
    case ABSY:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad|=comp->getmem(pc++)<<8;
      ad2=ad+y;
      if ((ad2&0xff00)!=(ad&0xff00))
        cyclesSingle++;
      comp->setmem(ad2,val);
      return;
    case ZP:
      cyclesSingle+=3;
      ad=comp->getmem(pc++);
      comp->setmem(ad,val);
      return;
    case ZPX:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad+=x;
      comp->setmem(ad&0xff,val);
      return;
    case ZPY:
      cyclesSingle+=4;
      ad=comp->getmem(pc++);
      ad+=y;
      comp->setmem(ad&0xff,val);
      return;
    case INDX:
      cyclesSingle+=6;
      ad=comp->getmem(pc++);
      ad+=x;
      ad2=comp->getmem(ad&0xff);
      ad++;
      ad2|=comp->getmem(ad&0xff)<<8;
      comp->setmem(ad2,val);
      return;
    case INDY:
      cyclesSingle+=5;
      ad=comp->getmem(pc++);
      ad2=comp->getmem(ad);
      ad2|=comp->getmem((ad+1)&0xff)<<8;
      ad=ad2+y;
      comp->setmem(ad,val);
      return;
    case ACC:
      cyclesSingle+=2;
      a=val;
      return;
    }
}


void MOS6510::status(FILE *f) const
{
  char sb[9];
  int bla=p;
  for (int i=0; i<=7; i++)
    {
      sb[i]=(bla&0x80)?BITtrue:BITfalse;
      bla<<=1;
    }
  sb[8]=0;
  fprintf(f, "%02x %02x %02x %02x %s ", a, x, y, s, sb);

  disass(pc, f);
}


const int MOS6510::disass(int loc, FILE *f, const bool newline) const 
{
  if(loc<0 || loc>0xFFFF)
    loc=pc;

  fprintf(f, (debug && breakpointlist.find(pc)!=breakpointlist.end())?"BP ":"   ");

  const byte opc=comp->getmem(loc);
  const int cmd=opcodes[opc];
  const int addr=modes[opc];

  fprintf(f, "%04x: %s %s",loc, disOP[cmd],disADR1[addr]);
  const int lll=disADRl[addr];
  if(addr==REL)
    {
      int r=comp->getmem(loc+1);
      if(r>127) r-=256;
      fprintf(f, "$%04x", loc+1+lll+r);
    }
  else
    {
      if (lll)
	printf("$");
      for (int xxx=lll; xxx>0; xxx--)
	fprintf(f, "%02x",comp->getmem(loc+xxx));
      fprintf(f, "%s",disADR2[addr]);
    }

  if(newline)
    fputc('\n', f);

  return loc+1+lll;
}

void MOS6510::reset(const int npc, const byte na, const byte nx, const byte ny)
{
  if(npc<0 || npc>0xFFFF)
    pc=lohi(comp->getmem(0xfffc), comp->getmem(0xfffd));
  else
    pc=static_cast<word>(npc);
  a=na;
  x=nx;
  y=ny;
  p=0;
  s=255;

  cycles=cyclesTotal=instructions=instructionsTotal=0;
}

void MOS6510::info(FILE *f) const
{
  fprintf(f, "ar xr yr sp nv-bdizc    pc    opcode\n");
  status(f);
}

const int MOS6510::parse()
{
  cyclesSingle=0; // reset cycle counter for this instruction

  // get opcode and adress mode
  const int cmd=opcodes[comp->getmem(pc)];
  int addr=modes[comp->getmem(pc)];

  pc++; // go to the operand byte
  int c; // need this in the rotate/shift intructions
  word wval;
  byte bval;
  switch (cmd)
    {
    case ADC:
      bval=getaddr(addr);
      if(p & FLAG_D)
	{
	  // code ripped from 64doc
	  byte AL = (a & 15) + (bval & 15) + ((p&FLAG_C)?1:0); // Calculate the lower nybble.
	  byte AH = (a >> 4) + (bval >> 4) + (AL > 15); // Calculate the upper nybble.

	  if (AL > 9) AL += 6;                  // BCD fixup for lower nybble.

	  setflags(FLAG_Z, a + bval + ((p&FLAG_C)?1:0));

	  // Negative and Overflow flags are set with the same logic than in Binary mode, but after fixing the lower nybble.
	  setflags(FLAG_N, (AH & 8 != 0));
	  setflags(FLAG_V, ((AH << 4) ^ a) & 128 && !((a ^ bval) & 128));

	  if (AH > 9) AH += 6;                  // BCD fixup for upper nybble.

	  // Carry is the only flag set after fixing the result.
	  setflags(FLAG_C, (AH > 15));
	  a = ((AH << 4) | (AL & 15)) & 255;
	}
      else
	{
	  wval=static_cast<word>(a)+bval+((p&FLAG_C)?1:0);
	  setflags(FLAG_C, wval&0x100);
	  a=wval;
	  setflags(FLAG_Z, !a);
	  setflags(FLAG_N, a&0x80);
	  setflags(FLAG_V, (!!(p&FLAG_C)) ^ (!!(p&FLAG_N)));
	}
      break;
    case ALR:
      bval=a&getaddr(addr);
      addr=ACC;
      goto lsr_l;
    case ANC:
    case AND:
      a&=getaddr(addr);
      setflags(FLAG_Z, !a);
      setflags(FLAG_N, a&0x80);
      if(cmd==ANC)
	setflags(FLAG_C, a&0x80);
      break;
    case ARR:
      // does not check decimal mode! see vice
      bval=a&getaddr(addr);
      addr=ACC;
      goto ror_l;
    case ASO:
    case ASL:
      wval=getaddr(addr);
      wval<<=1;
      if(cmd==ASO)
	wval|=a;
      setaddr(addr,wval);
      setflags(FLAG_Z,!wval);
      setflags(FLAG_N,wval&0x80);
      setflags(FLAG_C,wval&0x100);
      break;
    case AXA:
      putaddr(addr, a&x&comp->getmem(pc+1));
      break;
    case AXS:
      putaddr(addr, a&x);
      break;
    case BCC:
      branch(!(p&FLAG_C));
      break;
    case BCS:
      branch(p&FLAG_C);
      break;
    case BNE:
      branch(!(p&FLAG_Z));
      break;
    case BEQ:
      branch(p&FLAG_Z);
      break;
    case BPL:
      branch(!(p&FLAG_N));
      break;
    case BMI:
      branch(p&FLAG_N);
      break;
    case BVC:
      branch(!(p&FLAG_V));
      break;
    case BVS:
      branch(p&FLAG_V);
      break;
    case BIT:
      bval=getaddr(addr);
      setflags(FLAG_Z,!(a&bval));
      setflags(FLAG_N,bval&0x80);
      setflags(FLAG_V,bval&0x40);
      break;
    case BRK:
      pc++;
      push(pc>>8);
      push(pc&0xff);
      setflags(FLAG_B,1);
      push(p);
      pc=comp->getmem(0xfffe)|(comp->getmem(0xffff)<<8);
      cyclesSingle+=7;
      break;
    case CLC:
      cyclesSingle+=2;
      setflags(FLAG_C,0);
      break;
    case CLD:
      cyclesSingle+=2;
      setflags(FLAG_D,0);
      break;
    case CLI:
      cyclesSingle+=2;
      setflags(FLAG_I,0);
      break;
    case CLV:
      cyclesSingle+=2;
      setflags(FLAG_V,0);
      break;
    case CMP:
      bval=getaddr(addr);
    cmp_l:
      wval=static_cast<word>(a)-bval;
      setflags(FLAG_Z,!wval);
      setflags(FLAG_N,wval&0x80);
      setflags(FLAG_C,a>=bval);
      break;
    case CPX:
      bval=getaddr(addr);
      wval=static_cast<word>(x)-bval;
      setflags(FLAG_Z,!wval);
      setflags(FLAG_N,wval&0x80);
      setflags(FLAG_C,x>=bval);
      break;
    case CPY:
      bval=getaddr(addr);
      wval=static_cast<word>(y)-bval;
      setflags(FLAG_Z,!wval);
      setflags(FLAG_N,wval&0x80);
      setflags(FLAG_C,y>=bval);
      break;
    case DCM:
    case DEC:
      bval=getaddr(addr);
      bval--;
      setaddr(addr,bval);
      if(cmd==DCM)
	goto cmp_l;
      setflags(FLAG_Z,!bval);
      setflags(FLAG_N,bval&0x80);
      break;
    case DEX:
      cyclesSingle+=2;
      x--;
      setflags(FLAG_Z,!x);
      setflags(FLAG_N,x&0x80);
      break;
    case DEY:
      cyclesSingle+=2;
      y--;
      setflags(FLAG_Z,!y);
      setflags(FLAG_N,y&0x80);
      break;
    case EOR:
      bval=getaddr(addr);
      a^=bval;
      setflags(FLAG_Z,!a);
      setflags(FLAG_N,a&0x80);
      break;
    case HLT:
      break;
    case INS:
    case INC:
      bval=getaddr(addr);
      bval++;
      if(cmd==INS)
	goto sbc_l;
      setaddr(addr,bval);
      setflags(FLAG_Z,!bval);
      setflags(FLAG_N,bval&0x80);
      break;
    case INX:
      cyclesSingle+=2;
      x++;
      setflags(FLAG_Z,!x);
      setflags(FLAG_N,x&0x80);
      break;
    case INY:
      cyclesSingle+=2;
      y++;
      setflags(FLAG_Z,!y);
      setflags(FLAG_N,y&0x80);
      break;
    case JMP:
      cyclesSingle+=3;
      wval=comp->getmem(pc++);
      wval|=comp->getmem(pc++)<<8;
      switch(addr)
	{
        case ABS:
          pc=wval;
          break;
        case IND:
          pc=comp->getmem(wval);
          pc|=comp->getmem(wval+1)<<8;
          cyclesSingle+=2;
          break;
	}
      break;
    case JSR:
      bval=comp->getmem(pc++);
      push(pc>>8);
      push(pc&0xff);
      pc=bval|(comp->getmem(pc)<<8);
      cyclesSingle+=6;
      break;
    case LAS:
      a=x=s=s&getaddr(addr);
      setflags(FLAG_Z,!a);
      setflags(FLAG_N,a&0x80);
      break;
    case LAX:
    case LDA:
      a=getaddr(addr);
      if(cmd==LAX)
	x=a;
      setflags(FLAG_Z,!a);
      setflags(FLAG_N,a&0x80);
      break;
    case LDX:
      x=getaddr(addr);
      setflags(FLAG_Z,!x);
      setflags(FLAG_N,x&0x80);
      break;
    case LDY:
      y=getaddr(addr);
      setflags(FLAG_Z,!y);
      setflags(FLAG_N,y&0x80);
      break;
    case LSE:
    case LSR:
      bval=getaddr(addr);
    lsr_l:
      wval=bval;
      wval>>=1;
      if(cmd==LSE)
	wval^=a;
      setaddr(addr,wval);
      setflags(FLAG_Z,!wval);
      setflags(FLAG_N,wval&0x80);
      setflags(FLAG_C,bval&1);
      break;
    case NOP:
      cyclesSingle+=2;
      break;
    case OAL:
      a|=0xEE;
      a&=getaddr(addr);
      x=a;
      setflags(FLAG_Z,!a);
      setflags(FLAG_N,a&0x80);
      break;
    case ORA:
      bval=getaddr(addr);
      a|=bval;
      setflags(FLAG_Z,!a);
      setflags(FLAG_N,a&0x80);
      break;
    case PHA:
      push(a);
      cyclesSingle+=3;
      break;
    case PHP:
      push(p);
      cyclesSingle+=3;
      break;
    case PLA:
      a=pop();
      setflags(FLAG_Z,!a);
      setflags(FLAG_N,a&0x80);
      cyclesSingle+=4;
      break;
    case PLP:
      p=pop();
      cyclesSingle+=4;
      break;
    case RLA:
    case ROL:
      bval=getaddr(addr);
      c=!!(p&FLAG_C);
      setflags(FLAG_C,bval&0x80);
      bval<<=1;
      bval|=c;
      if(cmd==RLA)
	bval|=a;
      setaddr(addr,bval);
      setflags(FLAG_N,bval&0x80);
      setflags(FLAG_Z,!bval);
      break;
    case RRA:
    case ROR:
      bval=getaddr(addr);
    ror_l:
      c=!!(p&FLAG_C);
      setflags(FLAG_C,bval&1);
      bval>>=1;
      bval|=128*c;
      if(cmd==RRA)
	bval+=a+c;
      setaddr(addr,bval);
      setflags(FLAG_N,bval&0x80);
      setflags(FLAG_Z,!bval);
      break;
    case RTI:
      p=pop();
    case RTS:
      pc=pop();
      pc|=pop()<<8;
      if(cmd==RTS)
	pc++;
      cyclesSingle+=6;
      break;
    case SAX:
      x&=a;
      x-=getaddr(addr);
      setflags(FLAG_Z,!x);
      setflags(FLAG_N,x&0x80);
      setflags(FLAG_C,a>=x);
      /* flags? */
      break;
    case SAY:
      putaddr(addr, y&comp->getmem(pc+1));
      /* flags? */
      break;
    case SBC:
      bval=getaddr(addr);
    sbc_l:
      if(p&FLAG_D)
	{
	  // code ripped from 64doc
	  const byte C=((p&FLAG_C)?1:0);

	  byte AL = (a & 15) - (bval & 15) - !C; // Calculate the lower nybble
	  if (AL & 16) AL -= 6;	// BCD fixup for lower nybble

	  byte AH = (a >> 4) - (bval >> 4) - (AL & 16); // Calculate the upper nybble
	  if (AH & 16) AH -= 6;	// BCD fixup for upper nybble

	  // The flags are set just like in Binary mode
	  const byte b=a - bval - !C;
	  setflags(FLAG_C, b & 0x100);
	  setflags(FLAG_Z, b & 0xff);
	  setflags(FLAG_V, (b ^ bval) & 128 && (a ^ bval) & 128);
	  setflags(FLAG_N, b & 128);

	  a = ((AH << 4) | (AL & 15)) & 255;
	}
      else
	{
	  bval^=0xff;
	  wval=static_cast<word>(a)+bval+((p&FLAG_C)?1:0);
	  setflags(FLAG_C, wval&0x100);
	  a=wval;
	  setflags(FLAG_Z, !a);
	  setflags(FLAG_N, a>127);
	  setflags(FLAG_V, (!!(p&FLAG_C)) ^ (!!(p&FLAG_N)));
	}
      break;
    case SEC:
      cyclesSingle+=2;
      setflags(FLAG_C,1);
      break;
    case SED:
      cyclesSingle+=2;
      setflags(FLAG_D,1);
      break;
    case SEI:
      cyclesSingle+=2;
      setflags(FLAG_I,1);
      break;
    case SKB:
    case SKW:
      getaddr(addr);
      break;
    case STA:
      putaddr(addr,a);
      break;
    case STX:
      putaddr(addr,x);
      break;
    case STY:
      putaddr(addr,y);
      break;
    case TAS:
      /* correct? */
      bval=s=a&x;
      bval&=comp->getmem(pc+1) + 1;
      putaddr(addr, bval);
      break;
    case TAX:
      cyclesSingle+=2;
      x=a;
      setflags(FLAG_Z, !x);
      setflags(FLAG_N, x&0x80);
      break;
    case TAY:
      cyclesSingle+=2;
      y=a;
      setflags(FLAG_Z, !y);
      setflags(FLAG_N, y&0x80);
      break;
    case TSX:
      cyclesSingle+=2;
      x=s;
      setflags(FLAG_Z, !x);
      setflags(FLAG_N, x&0x80);
      break;
    case TXA:
      cyclesSingle+=2;
      a=x;
      setflags(FLAG_Z, !a);
      setflags(FLAG_N, a&0x80);
      break;
    case TXS:
      cyclesSingle+=2;
      s=x;
      break;
    case TYA:
      cyclesSingle+=2;
      a=y;
      setflags(FLAG_Z, !a);
      setflags(FLAG_N, a&0x80);
      break;
    case XAA:
#if 1
      a=(a|0xEE)&x&getaddr(addr);
      setflags(FLAG_Z, !a);
      setflags(FLAG_N, a&0x80);
#else
      a=x&getaddr(addr);
      /* flags? */
#endif
      break;
    case XAS:
      putaddr(addr, x&comp->getmem(pc+1));
      break;
    }

  cycles+=cyclesSingle;
  cyclesTotal+=cyclesSingle;
  instructions++;
  instructionsTotal++;

  switch(cmd)
    {
    case BRK:
      infostr.insert("CPU halted by BRK");
      return -cycles;

    case HLT:
      infostr.insert("CPU halted by HLT");
      return -cycles;

    case ASO: infostr.insert("illegal opcode ASO"); break;
    case RLA: infostr.insert("illegal opcode RLA"); break;
    case LSE: infostr.insert("illegal opcode LSE"); break;
    case RRA: infostr.insert("illegal opcode RRA"); break;
    case AXS: infostr.insert("illegal opcode AXS"); break;
    case LAX: infostr.insert("illegal opcode LAX"); break;
    case DCM: infostr.insert("illegal opcode DCM"); break;
    case INS: infostr.insert("illegal opcode INS"); break;
    case ALR: infostr.insert("illegal opcode ALR"); break;
    case ARR: infostr.insert("illegal opcode ARR"); break;
    case XAA: infostr.insert("illegal opcode XAA"); break;
    case OAL: infostr.insert("illegal opcode OAL"); break;
    case SAX: infostr.insert("illegal opcode SAX"); break;
    case SKB: infostr.insert("illegal opcode SKB"); break;
    case SKW: infostr.insert("illegal opcode SKW"); break;
    case TAS: infostr.insert("illegal opcode TAS"); break;
    case SAY: infostr.insert("illegal opcode SAY"); break;
    case XAS: infostr.insert("illegal opcode XAS"); break;
    case AXA: infostr.insert("illegal opcode AXA"); break;
    case ANC: infostr.insert("illegal opcode ANC"); break;
    case LAS: infostr.insert("illegal opcode LAS"); break;
    }

  return cyclesSingle;
}

const int MOS6510::jsr(const int npc, const int na, const int nx, const int ny, const int maxCycles)
{
  if(npc>=0 && npc<=0xFFFF)
    {
      pc=npc;
      p=0;
      s=255;
      // RTS trick
      push(0xFF);
      push(0xFF);
    }
  else if(npc==-2)
    {
      pc=lohi(comp->getmem(0xFFFC), comp->getmem(0xFFFD));
      p=0;
      s=255;
      // RTS trick
      push(0xFF);
      push(0xFF);
    }

  if(na>=0 && na<=0xFF)
    a=na;
  if(nx>=0 && nx<=0xFF)
    x=nx;
  if(ny>=0 && ny<=0xFF)
    y=ny;

  cycles=instructions=0;

  abort=false;
  do
    {
      // check max cycles
      if(maxCycles>0 && cycles>maxCycles)
	{
	  infostr.insert("CPU reached maxCycles");
	  break;
	}
      // check breakpoint
      if(debug && breakpoint(pc))
	break;
      // check BRK,HLT
      if(parse()<0)
	break;
      // check RTS trick
      if(pc==0)
	break;
      // check abort
      if(debug && abort)
	break;
    }
  while(true);

  return cycles;
}

void MOS6510::push(const byte val)
{
  comp->setmem(0x100+s, val);
  if(s)
    s--;
  else
    infostr.insert("stack overrun");
}

const byte MOS6510::pop()
{
  if(s<0xff)
    s++;
  else
    infostr.insert("stack underrun");
  return comp->getmem(0x100+s);
}

void MOS6510::branch(const bool flag)
{
  // get branch adress
  word w=pc+(signed char)getaddr(IMM);
  if (flag)
    {
      // calcute page border penalty
      cyclesSingle+=((pc&0x100)!=(w&0x100))?2:1;
      // branch
      pc=w;
    }
}

bool MOS6510::breakpoint(int addr)
{
  return breakpointlist.find(addr)!=breakpointlist.end();
}

void MOS6510::breakpointToggle(int addr)
{
  if(addr<0 || addr>0xFFFF)
    addr=pc;

  if(breakpoint(addr))
    {
      breakpointlist.erase(addr);
      printf("clearing BP at %04X\n", addr);
    }
  else
    {
      breakpointlist.insert(addr);
      printf("setting BP at %04X\n", addr);
    }
}

void MOS6510::breakpointPrint()
{
  printf("BreakPoints: ");
  for(breakpointlist_t::iterator i=breakpointlist.begin(); i!=breakpointlist.end(); ++i)
    printf("%04X ", *i);
  printf("\n");
}
