/*
  sid (c) 1999-2004 by Dirk Jagdmann <doj@cubic.org>

  This file is part of sid.

  sid is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  sid is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with sid; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  
  $Header: /code/cw/sid/psid.h,v 1.6 2004/05/14 14:51:04 doj Exp $
*/

#ifndef PSID__H
#define PSID__H

#include <stdio.h>

#include "defines.h"

#pragma pack(push, 1)

/**
   structure describing a sid file. The first $78/$7C bytes are identical to the [PR]SID fileformat header. Look at the PSID docs for the undocumented variables in this struct.
*/
struct psid {
  char id[4];
  word version;
  word dataOffset;
  word loadAddr;
  word initAddr;
  word playAddr;
  word songTotal;
  word songStart;
  dword speed;
  char name[32];
  char author[32];
  char copyright[32];

  word flags;
  byte startPage;
  byte pageLength;
  word reserved;

  // PSID header ends here!

  /// currently played subtune
  word songPlay;
  /// song is in .mus format
  bool MUSplayer;
  /// song contains PlaySID samples
  bool PlaySIDSamples;
  /// song should be played with 6581
  bool mos6581;
  /// song should be played with 8580
  bool mos8580;
  /// song should be played with PAL/NTSC frequency
  bool pal;
};

#pragma pack(pop)

/// @return true if f contains a valid PSID file
bool isPSID(FILE *f);

/*
  load a PSID from f into mem.

  @param f stdio file
  @param mem pointer to 64K memory block
  @return pointer to psid structure filled with valid values. If no file could be loaded 0 is returned. The caller must delete the returned struct.
*/
psid* loadPSID(FILE *f, byte *mem);

#endif
